#  %W%  %G% CSS
#  "pyspec" Release %R%
#
#$Id: SpecChannel.py,v 1.6 2006/12/14 10:03:13 guijarro Exp $
"""SpecChannel module

This module defines the SpecChannel class
"""

__author__ = 'Matias Guijarro'
__version__ = '1.0'

import SpecEventsDispatcher  as SpecEventsDispatcher
from SpecWaitObject import SpecWaitObject
from SpecClientError import SpecClientError
import weakref

from pyspec.css_logger import log

(DOREG, DONTREG, WAITREG) = (0, 1, 2)

class SpecChannel(object):
    """SpecChannel class

    Represent a channel in Spec

    Signals:
    valueChanged(value, chan_name) -- emitted when the channel gets updated
    """

    def __init__(self, conn, chan_name, registrationFlag = DOREG):
        """Constructor

        Arguments:
        * conn -- a SpecConnection object
        * chan_name -- string representing a channel name, i.e. 'var/toto'

        Keyword arguments:
        * registrationFlag -- defines how the channel is registered, possible
             values are : SpecChannel.DOREG (default), SpecChannel.DONTREG
             (do not register), SpecChannel.WAITREG (delayed registration until Spec is
             reconnected)
        """
        self.conn = weakref.ref(conn)
        self.name = chan_name

        if chan_name.startswith("var/") and '/' in chan_name[4:]:
            l = chan_name.split('/')
            self.spec_chan_name = "/".join((l[0], l[1]))

            if len(l)==3:
                self.access1=l[2]
                self.access2=None
            else:
                self.access1=l[2]
                self.access2=l[3]
        else:
            self.spec_chan_name = self.name
            self.access1=None
            self.access2=None

        self.registrationFlag = registrationFlag

        self._connected = False
        self.registered = False
        self.value = None

        SpecEventsDispatcher.connect(conn, 'connected', self.connected)
        SpecEventsDispatcher.connect(conn, 'disconnected', self.disconnected)

        if conn.is_connected():
            self.connected()

    def connected(self):
        """Do registration when Spec gets connected

        If registration flag is WAITREG put the flag to DOREG if not yet connected,
        and register if DOREG
        """
        if self.registrationFlag == WAITREG:
            if not self._connected:
                self.registrationFlag = DOREG

        self._connected = True

        if self.registrationFlag == DOREG:
            self.register()

    def disconnected(self):
        """Reset channel object when Spec gets disconnected."""
        self.value = None
        self._connected = False

    def read(self):
        """Read the channel value

        If channel is registered, just return the internal value,
        else obtain the channel value and return it.
        """
        if self.registered and self.value is not None:
            return self.value
        
        conn = self.conn()

        if conn is not None:
            #w = SpecWaitObject(conn)
            # make sure spec is connected, we give a short timeout
            # because it is supposed to be the case already
            #w.waitConnection(timeout=500)                                 
            #w.waitReply('send_msg_chan_read', (self.spec_chan_name, ))
            #self.update(w.value)

            conn.wait_connected()

            reply_id = conn.send_msg_chan_read(self.spec_chan_name)

            try:
                self.value = conn.wait_reply(reply_id)
            except SpecClientError as e:
                raise(e)

        return self.value

    def write(self, value):
        """Write a channel value."""
        conn = self.conn()

        if conn is not None:
            if self.access1 is not None:
                if self.access2 is None:
                    value = { self.access1: value }
                else:
                    value = { self.access1: { self.access2: value } }

            conn.send_msg_chan_send(self.spec_chan_name, value)

    def register(self):
        """Register channel

        Registering a channel means telling the server we want to receive
        update events when a channel value changes on the server side.
        """
        if self.spec_chan_name != self.name:
            return

        conn = self.conn()

        if conn is not None:
            conn.send_msg_register(self.spec_chan_name)
            self.registered = True

    def update(self, value, deleted = False,force=False):
        """Update channel's value and emit the 'valueChanged' signal."""

        # receive dictionary - access1 is set (init with var/arr/idx1 or var/arr/idx1/idx2 )
        if isinstance(value, dict) and self.access1 is not None:
            if self.access1 in value:
                if deleted:
                    SpecEventsDispatcher.emit(self, 'valueChanged', (None, self.name, ))
                else:
                    if self.access2 is None:
                        if force or self.value is None or self.value != value[self.access1]: 
                            self.value = value[self.access1]
                            SpecEventsDispatcher.emit(self, 'valueChanged', (self.value, self.name, ))
                    else:
                        if self.access2 in value[self.access1]:
                            if deleted:
                                SpecEventsDispatcher.emit(self, 'valueChanged', (None, self.name, ))
                            else:
                                if force or self.value is None or self.value != value[self.access1][self.access2]:
                                    self.value = value[self.access1][self.access2]
                                    SpecEventsDispatcher.emit(self, 'valueChanged', (self.value, self.name, ))
            return

        # receive dictionary - access1 is not set (init with var/arr)
        if isinstance(self.value, dict) and isinstance(value, dict):
            # update dictionary
            if deleted:
                for key,val in iter(value.items()):
                    if isinstance(val,dict):
                        for k in val:
                            try:
                                del self.value[key][k]
                            except KeyError:
                                pass
                        if len(self.value[key])==1 and None in self.value[key]:
                            self.value[key]=self.value[key][None]
                    else:
                        try:
                            del self.value[key]
                        except KeyError:
                            pass
            else:
                for k1,v1 in iter(value.items()):
                    if isinstance(v1,dict):
                        try:
                            self.value[k1].update(v1)
                        except KeyError:
                            self.value[k1]=v1
                        except AttributeError:
                            self.value[k1]={None: self.value[k1]}
                            self.value[k1].update(v1)
                    else:
                        if k1 in self.value.keys() and isinstance(self.value[k1], dict):
                            self.value[k1][None] = v1
                        else:
                            self.value[k1] = v1
            value2emit=self.value.copy()
            SpecEventsDispatcher.emit(self, 'valueChanged', (value2emit, self.name, ))
            return

        # no assoc array - (init with var/var_name)
        if deleted:
            self.value = None
        else:
            self.value = value
            SpecEventsDispatcher.emit(self, 'valueChanged', (self.value, self.name, ))

    def unregister(self):
        """Unregister channel."""
        conn = self.conn()

        if conn is not None:
            conn.send_msg_unregister(self.spec_chan_name)
            self.registered = False
            self.value = None



#!/usr/bin/env python

#******************************************************************************
#
#  %W%  %G% CSS
#
#  "pyspec" Release %R%
#
#  Copyright (c) 2013,2014,2015,2016,2018,2020
#  by Certified Scientific Software.
#  All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software ("pyspec") and associated documentation files (the
#  "Software"), to deal in the Software without restriction, including
#  without limitation the rights to use, copy, modify, merge, publish,
#  distribute, sublicense, and/or sell copies of the Software, and to
#  permit persons to whom the Software is furnished to do so, subject to
#  the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  Neither the name of the copyright holder nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
#
#     * The software is provided "as is", without warranty of any   *
#     * kind, express or implied, including but not limited to the  *
#     * warranties of merchantability, fitness for a particular     *
#     * purpose and noninfringement.  In no event shall the authors *
#     * or copyright holders be liable for any claim, damages or    *
#     * other liability, whether in an action of contract, tort     *
#     * or otherwise, arising from, out of or in connection with    *
#     * the software or the use of other dealings in the software.  *
#
#******************************************************************************

"""

****************
specfile
****************

Description
****************
Utility to handle data files in spec data format

Classes
****************
To check the definition of the spec data file format check `certif.com`_

.. _`certif.com`: http://www.certif.com/spec_help/scans.html

"""

version = "1.0"

import sys
import os
import getopt

SPECD='/usr/local/lib/spec.d'
specd = os.environ.get('SPECD', SPECD)
sys.path.append( specd )

try:
   from pyspec.file.spec import FileSpec
except ImportError:
   print("Cannot find filespec module. Try setting SPECD variable")
   sys.exit(0)

outformats = ['csv', 'tabs', 'spec']

def printUsage(msg=None, longmode=False):
    if msg:
       print(msg)

    if not longmode:
       print("""Usage: %(progname)s [options] filename [scanlist] 
   type \"%(progname)s -h\" for a detailed help """ % {'progname': os.path.basename(sys.argv[0])} )
    else:
       print("""
Usage: %(progname)s [options] filename [scanlist]

Options are: 
  -f format
      Format of the output files. Format can be one out of "tabs", "csv" or "spec"
      Default output format is "tabs"

  -O 
      Do not overwrite existing files. By default %(progname)s will overwrite existing files
      if they have the same name

  -S 
      Use single file for output. Useful particularly in the case of "spec" output format

  -p  prefix
      Use `prefix` as prefix for output files
      The default is to set prefix as the the root name of the original file

  -s  suffix

  -d outdir
      Use `outdir` directory to write output files. The default is to use the  
      root name of the original file. 
      
  -l 
      List scans in file in condensed mode
  -L 
      List scans in file in non condensed mode

  -a 
      Extract all scans in the file

  -h 
      Prints this help and exits

  -V 
      Prints program version


Specifying a scanlist:
  scanlist can be given as a list of scan numbers or scan ranges (
  using ":" to specify a specify first and last scan number in range. 
  Either "," or <space> can be used 
  to separate arguments.

  Examples of valid scan specification are:
     %(progname)s myfile.dat 3 6 8
     %(progname)s myfile.dat 3,6,8
     %(progname)s myfile.dat 3,6:12,37
     %(progname)s myfile.dat 3 6:12 37
     %(progname)s myfile.dat 3,6:12 37
 
  Remember you can also use the "-a" flag to extract all scans in a file


""" % { 'progname': os.path.basename(sys.argv[0])})

def parseScanArgs(scanarg):

    args1 = scanarg.split()
    args = [] 
    for arg in args1:
       args.extend( arg.split(",") )

    scanlist = []

    for arg in args:
      if arg.find(":") != -1:
         slist = None
         try:
             arg1, arg2 = arg.split(":")
             iarg1 = int(arg1)
             iarg2 = int(arg2)
             if iarg2 <= iarg1:
                raise "Bad scan arguments"
             else:
                slist = [ str(val) for val in range(iarg1, iarg2+1) ]
         except:
             raise "Bad scan arguments"
         scanlist.extend(slist)
      else:
         iarg = arg
         scanlist.append(iarg)

    return scanlist

def formatScanList(scanlist, condensed=True):
    strlist = ""
    prev = -1
    openlist = False

    sep = ""
    for scanno in scanlist:
        if not condensed: 
           strlist += ( sep + str(scanno) )
        else:
           if scanno == (prev + 1):
              openlist = True
           else:
              if openlist:
                 strlist += ":"
                 strlist += str(prev)
              strlist += ( sep + str(scanno) )
              openlist = False
        prev = scanno
        sep = ","
    else:
        if condensed and openlist:
           strlist += ":"
           strlist += str(scanno)

    return strlist

def main():

    outformat = "tabs"
    suffix    = None
    prefix    = None
    outdir    = None

    overw_flag  = False
    single_flag = False
    all_flag    = False
    list_flag   = False

    if len(sys.argv) < 2:
       printUsage()
       sys.exit(0)

    try:
       optlist, args = getopt.getopt(sys.argv[1:], "f:s:p:d:alLOShV")
    except:
       printUsage(msg="wrong usage")
       sys.exit(1)

    for o,a in optlist:
        if o == '-h':
             printUsage(longmode=True)
             sys.exit(0)
        elif o == '-V':
            print(version)
            sys.exit(0)
        elif o == "-f":
            outformat = a
        elif o == '-p':
            prefix = a
        elif o == '-s':
            suffix = a
        elif o == '-d':
            outdir = a
        elif o == "-l":
            list_flag = True
            condensed = True
        elif o == "-L":
            list_flag = True
            condensed = False
        elif o == '-O':
            overw_flag = True
        elif o == '-a':
            all_flag = True
        elif o == '-S':
            single_flag = True

    if len(args) == 0:
       printUsage("You should specify an input filename")
       sys.exit(0)

    if outformat not in outformats:
        printUsage( msg="Wrong output format specified. Valid formats are %s" % ",".join(outformats))
        sys.exit(1)

    filename = args[0]

    if len(args) > 1:
       scanargs = " ".join( args[1:] )
    else:
       scanargs = None

    # check if file exists and it is plain and readable file
    if not os.path.exists(filename):
        print("File %s does not exist." % filename) 
        sys.exit(1)

    # open file. check if any scan could  be indexed
    try:
        fs = FileSpec(filename)
    except:
        import traceback
        traceback.print_exc()
        print("Cannot index file %s." % filename)
        sys.exit(1)

    if len(fs) <= 0:
        print("Cannot index file %s." % filename)
        sys.exit(1)

    if list_flag:
        scanlist = [ scan.getNumber() for scan in fs ]
        strlist = formatScanList( scanlist, condensed=condensed )
        print(strlist)
        sys.exit(0)

    # prepare the scan list to extract
    scanlist = None
    if scanargs is not None:
       #try: 
          scanlist = parseScanArgs( scanargs ) 
       #except:
          #print("Wrong scan selection")
          #sys.exit(1)
       
    scans = []
    if scanlist:
       for scanno in scanlist:
           sparts = scanno.split(".")
           sno  = int(sparts[0])
           if len(sparts) > 1:
              sord = int(sparts[1])
           else:
              sord = 0
           scan = fs.getScanByNumber( int(sno), int(sord) ) 
           if scan:
              scans.append( scan ) 
           else:
              print("Cannot find scan %d(%d) in file %s" % (sno,sord,filename))
    else:
       if all_flag:
          # extract them all
          scans = [ scan for scan in fs ]

    inprefix = os.path.splitext( os.path.basename( filename ))[0]

    if not outdir:
       outdir = inprefix

    if not suffix:
       if outformat == "csv":
          suffix = "csv"
       else:
          suffix = "dat"

    if not prefix:
        prefix = inprefix

    if not os.path.exists(outdir):
       os.makedirs(outdir)

    for scan in scans:
 
       if single_flag:
          outfile = os.path.join( outdir, "%s_bundle.%s" % ( prefix, suffix ))
       else:
          outfile = os.path.join( outdir, "%s_%s.%s" % ( prefix, scan.getNumber(), suffix ))

       # find alternative name if not set to overwrite
       if not overw_flag and not single_flag:
          tryno = 0
          while os.path.exists(outfile):
             tryno += 1
             outfile = os.path.join( outdir, "%s_%s-%d.%s" % ( prefix, scan.getNumber(), tryno, suffix )) 

       scan.save(outfile, format=outformat, append=single_flag)
 
if __name__ == "__main__":
     main()
